/*
*  MATRIX routines:
*    CLASSES Matrix,Stolb. (Only Squared Matrixes.)
*      A1,A2, ... - Matrixes.
*      S1,S2, ... - Stolbes.
*    Matrix operations:
*      A1=A2+A3
*      A1=A2-A3
*      A1=A2*A3    - now absent.
*      S1=S1+S2
*      S1=S1-S2
*      S1=A1*S2
*    Solutions of linear systems:
*      S2=Solve(A1,S1); - Solves system A1*S2=S1. This function
                          not changes A1 and S1.
*    Triangle:
*      A2=Triangle(A1); - Triagonalization of matrix.
*    Norma:
*      double Norma(S1); - sqrt(S1.x[1]^2+...).
*
*/
/************* Main definitions. ************************/

#include <iostream.h>
#include <math.h>

class Matrix
{
  public:
    Matrix();
    Matrix(int Nx1,int Ny1);
    Matrix(const Matrix& m1);
    ~Matrix();
    double Get(int i,int j);
    void Set(int i,int j,double m1);
    Matrix& operator=(Matrix& A1);
    friend Matrix operator+(Matrix& A1,Matrix& A2);
  double* Aa;
  int Nx,Ny;
};
class Stolb
{
  public:
    Stolb();
    Stolb(int Ns1);
    Stolb(const Stolb& m1);
    ~Stolb();
    double Get(int i);
    void Set(int i,double m1);
    Stolb& Stolb::operator=(Stolb& S1);
  double* As;
  int Ns;
};

/******* Functions bodies. *********************************/
Matrix& Matrix::operator=(Matrix& A1)
{
  Nx=A1.Nx;Ny=A1.Ny;
  for(int i=0;i<Nx*Ny;i++)
  {Aa[i]=A1.Aa[i];};
  return *this;
};
Stolb& Stolb::operator=(Stolb& S1)
{
  Ns=S1.Ns;
  for(int i=0;i<Ns;i++)
  {As[i]=S1.As[i];};
  return *this;
};
Matrix::Matrix(const Matrix& m1)
{
  Nx=m1.Nx;Ny=m1.Ny;
  Aa=new double[Nx*Ny];
  for(int i=0;i<Nx*Ny;i++)
    Aa[i]=m1.Aa[i];
};
Matrix::Matrix(int Nx1,int Ny1)
{
  Nx=Nx1;Ny=Ny1;
  Aa=new double[Nx*Ny];
  for(int i=0;i<Nx;i++)
  {for(int j=0;j<Ny;j++){
     if(i==j) Aa[i*Nx+j]=1.0; else Aa[i*Nx+j]=0.0;
  }};
};
Matrix::~Matrix()
{
  delete Aa;
};
void Print(Matrix& A1)
{
  cout<<"\n";
  for(int i=0;i<A1.Ny;i++)
  {for(int j=0;j<A1.Nx;j++)
  {
    cout<<A1.Aa[i*A1.Ny+j]<<"  ";
  }cout<<"\n";};
};
double Matrix::Get(int i,int j)
{
  return(Aa[i*Ny+j]);
};
void Matrix::Set(int i,int j,double m1)
{
  Aa[i*Ny+j]=m1;
};
Stolb::Stolb(int Ns1)
{
  Ns=Ns1;
  As=new double[Ns];
  for(int i=0;i<Ns;i++)
  {As[i]=0.0;};
};
Stolb::Stolb(const Stolb& m1)
{
  Ns=m1.Ns;
  As=new double[Ns];
  for(int i=0;i<Ns;i++)
    As[i]=m1.As[i];
};
Stolb::~Stolb()
{
  delete As;
};
double Stolb::Get(int i)
{
  return As[i];
};
void Stolb::Set(int i,double m1)
{
  As[i]=m1;
};
void  Print(Stolb& S1)
{
  cout<<"\n";
  for(int i=0;i<S1.Ns;i++)
  {cout<<S1.As[i]<<"\n";};
};
Stolb operator*(Matrix& A1,Stolb& S1)
{
  Stolb S2(S1.Ns);
  double m1=0;
  if(A1.Nx!=S1.Ns)
  {cout<<"\n Error 1: different dimensions of Matrix and Stolb.\n";return S2;}
  else
  {
   for(int i=0;i<A1.Ny;i++)
   {
     m1=0;
     for(int j=0;j<S1.Ns;j++)
     {
       m1=m1+S1.Get(j)*A1.Get(i,j);
     };
     S2.Set(i,m1);
   };
   return S2;
  };
};
Matrix Triangle(Matrix& Aq)
{
  int l,i,j,k;
  Matrix A3=Aq;
  double akk,aik;
  for(k=0;k<A3.Ny-1;k++)
  {
     if(A3.Aa[k*A3.Nx+k]==0)
     {
	l=k;
	while(A3.Aa[l*A3.Nx+k]==0)
	{
	  l++;
	};
	for(i=0;i<A3.Ny;i++)
	{
	  akk=A3.Aa[k*A3.Nx+i];
	  A3.Aa[k*A3.Nx+i]=A3.Aa[l*A3.Nx+i];
	  A3.Aa[l*A3.Nx+i]=akk;
	};
     };
     akk=A3.Aa[k*A3.Nx+k];
     for(i=k+1;i<A3.Nx;i++)
     {
       aik=A3.Aa[i*A3.Nx+k];
       if(aik)
       {
 //	 S1.As[i]=S1.As[i]/aik-S1.As[k]/akk;
	 for(j=0;j<A3.Ny;j++)
	 {
	   A3.Aa[i*A3.Nx+j]=A3.Aa[i*A3.Nx+j]/aik-A3.Aa[k*A3.Nx+j]/akk;
	 };
       };
     };
  };
  return A3;
};
void Triangle(Matrix& A3,Stolb& S1)
{
  int l,i,j,k;
  double akk,aik;
  for(k=0;k<A3.Ny-1;k++)
  {
     if(A3.Aa[k*A3.Nx+k]==0)
     {
	l=k;
	while(A3.Aa[l*A3.Nx+k]==0)
	{
	  l++;
	};
	for(i=0;i<A3.Ny;i++)
	{
	  akk=A3.Aa[k*A3.Nx+i];
	  A3.Aa[k*A3.Nx+i]=A3.Aa[l*A3.Nx+i];
	  A3.Aa[l*A3.Nx+i]=akk;
	};
     };
     akk=A3.Aa[k*A3.Nx+k];
     for(i=k+1;i<A3.Nx;i++)
     {
       aik=A3.Aa[i*A3.Nx+k];
       if(aik)
       {
	 S1.As[i]=S1.As[i]/aik-S1.As[k]/akk;
	 for(j=0;j<A3.Ny;j++)
	 {
	   A3.Aa[i*A3.Nx+j]=A3.Aa[i*A3.Nx+j]/aik-A3.Aa[k*A3.Nx+j]/akk;
	 };
       };
     };
  };
};
Stolb Solve(Matrix& Aq,Stolb& Bq)
{
  Stolb Y(Bq.Ns);
  double sum=0;
  Matrix A=Aq;Stolb B=Bq;
  Triangle(A,B);
  for(int i=B.Ns-1;i>=0;i--)
  {
    sum=0;
    for(int j=i+1;j<Y.Ns;j++)
    {
      sum=sum+A.Get(i,j)*Y.Get(j);
    };
    Y.Set(i,1.0/A.Get(i,i)*(B.Get(i)-sum));
  };
  return Y;
};
double Norma(Stolb& S1)
{
   double f=0;
   for(int i=0;i<S1.Ns;i++)
   {f=f+(S1.As[i])*(S1.As[i]);};
   return (f);
};
Matrix operator-(Matrix& A1,Matrix& A2)
{
  Matrix A3(A1.Ny,A1.Nx);
  for(int i=0;i<A1.Nx*A1.Ny;i++)
  {A3.Aa[i]=A1.Aa[i]-A2.Aa[i];};
  return A3;
};
Matrix operator+(Matrix& A1,Matrix& A2)
{
  Matrix A3(A1.Ny,A1.Nx);
  for(int i=0;i<A3.Nx*A3.Ny;i++)
    A3.Aa[i]=A1.Aa[i]+A2.Aa[i];
  return(A3);
};
Stolb operator-(Stolb& S1,Stolb& S2)
{
  Stolb S3(S1.Ns);
  for(int i=0;i<S1.Ns;i++)
  {S3.As[i]=S1.As[i]-S2.As[i];};
  return S3;
};
Stolb operator+(Stolb& S1,Stolb& S2)
{
  Stolb S3(S1.Ns);
  for(int i=0;i<S1.Ns;i++)
  {S3.As[i]=S1.As[i]+S2.As[i];};
  return S3;
};



